/*
 * Copyright 2006-2007 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.openspaces.maven.plugin;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.springframework.util.StringUtils;


/**
 * Goal which creates the openspaces project.
 */
@Mojo(name = "create", threadSafe = true, requiresProject = false)
@SuppressWarnings({"JavaDoc", "unused"})
public class CreatePUProjectMojo extends AbstractOpenSpacesMojo {

    /**
     * the groupId string to replace with the package name
     */
    private static final String FILTER_GROUP_ID = "${puGroupId}";


    /**
     * the groupId string to replace with the package name
     */
    private static final String FILTER_GROUP_PATH = "${puGroupPath}";


    /**
     * the groupId string to replace with the package name
     */
    private static final String FILTER_ARTIFACT_ID = "${puArtifactId}";


    /**
     * Templates directory name
     */
    private static final String DIR_TEMPLATES = "pu-templates";


    @Parameter(property = "artifactId", defaultValue = "my-app")
    private File projectDir;


    @Parameter(property = "groupId", defaultValue = "com.gigaspaces.app")
    private String packageName;


    @Parameter(property = "template")
    private String template;


    /**
     * The directory structure of the package.
     */
    @SuppressWarnings("FieldCanBeLocal")
    private String packageDirs;


    public void executeMojo() throws MojoExecutionException {
        try {
            List urls = getTemplatesURLs();
            ClassLoader cl = Utils.createClassLoader(urls, Thread.currentThread().getContextClassLoader());
            Thread.currentThread().setContextClassLoader(cl);

            if (template == null || template.trim().length() == 0) {
                //throw new IllegalArgumentException(createAvailableTemplatesMessage());
                System.out.println(createAvailableTemplatesMessage());
                return;
            }

            Enumeration templateUrl = Thread.currentThread().getContextClassLoader().getResources(DIR_TEMPLATES + "/" + template + "/");
            if (templateUrl.hasMoreElements()) {
                // template found
                URL url = (URL) templateUrl.nextElement();
                PluginLog.getLog().debug("Found template at: " + url);

                // extract the jar path from the url
                url = getJarURLFromURL(url, "/" + template + "/");
                PluginLog.getLog().debug("Template JAR url: " + url);

                extract(url);
            } else {
                // the template was not found - show available templates
                throw new IllegalArgumentException(createAvailableTemplatesMessage());
            }
        } catch (Exception e) {
            throw new MojoExecutionException("Failed to create processing unit project", e);
        }
    }


    /**
     * Returns a list that contains all extension templates URLs.
     */
    private List getTemplatesURLs() throws Exception {
        String pluginPath = getPluginPath();
        List<URL> urls = new ArrayList<URL>();
        if (pluginPath != null) {
            PluginLog.getLog().debug("Plugin path: " + pluginPath);
            File f = new File(pluginPath);
            File dir = f.getParentFile();
            File templatesDir = new File(dir, DIR_TEMPLATES);
            if (templatesDir.exists()) {
                File[] templates = templatesDir.listFiles();
                if (templates != null) {
                    for (File template1 : templates) {
                        //urls.add(Utils.getURL(templates[i]));
                        urls.add(template1.toURI().toURL());
                    }
                }
            }
        }
        return urls;
    }


    /**
     * Extracts the project files to the project directory.
     */
    private void extract(URL url) throws Exception {
        if (packageName.equals("com.gigaspaces.app")) {
            packageName = packageName + "." + template.replaceAll("-", "_");
        }
        packageDirs = packageName.replaceAll("\\.", "/");
        String puTemplate = DIR_TEMPLATES + "/" + template + "/";
        int length = puTemplate.length() - 1;
        BufferedInputStream bis = new BufferedInputStream(url.openStream());
        JarInputStream jis = new JarInputStream(bis);
        JarEntry je;
        byte[] buf = new byte[1024];
        int n;
        while ((je = jis.getNextJarEntry()) != null) {
            String jarEntryName = je.getName();
            PluginLog.getLog().debug("JAR entry: " + jarEntryName);
            if (je.isDirectory() || !jarEntryName.startsWith(puTemplate)) {
                continue;
            }
            String targetFileName = projectDir + jarEntryName.substring(length);

            // convert the ${gsGroupPath} to directory
            targetFileName = StringUtils.replace(targetFileName, FILTER_GROUP_PATH, packageDirs);
            PluginLog.getLog().debug("Extracting entry " + jarEntryName + " to " + targetFileName);

            // read the bytes to the buffer
            ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
            while ((n = jis.read(buf, 0, 1024)) > -1) {
                byteStream.write(buf, 0, n);
            }

            // replace property references with the syntax ${property_name}
            // to their respective property values.
            String data = byteStream.toString();
            data = StringUtils.replace(data, FILTER_GROUP_ID, packageName);
            data = StringUtils.replace(data, FILTER_ARTIFACT_ID, projectDir.getName());
            data = StringUtils.replace(data, FILTER_GROUP_PATH, packageDirs);

            // write the entire converted file content to the destination file.
            File f = new File(targetFileName);
            File dir = f.getParentFile();
            if (!dir.exists()) {
                //noinspection ResultOfMethodCallIgnored
                dir.mkdirs();
            }
            FileWriter writer = new FileWriter(f);
            writer.write(data);
            jis.closeEntry();
            writer.close();
        }
        jis.close();
    }


    /**
     * Returns the plugin path. Needs to be invoked from the main thread.
     *
     * @return the path to the plugin JAR file.
     */
    private String getPluginPath() throws Exception {
        Enumeration urls = Thread.currentThread().getContextClassLoader().getResources(DIR_TEMPLATES);
        while (urls.hasMoreElements()) {
            URL url = (URL) urls.nextElement();
            url = getJarURLFromURL(url, "");
            if (url.getProtocol().equals("file")) {
                return url.getPath();
            }
        }
        return null;
    }


    /**
     * Returns the path of a JAR file that appears in the URL
     */
    private URL getJarURLFromURL(URL url, String suffix) throws MalformedURLException {
        String urlString = url.toString();
        suffix = "!/" + DIR_TEMPLATES + suffix;
        if (urlString.endsWith(suffix)) {
            urlString = urlString.substring(4);
            urlString = urlString.substring(0, urlString.length() - suffix.length());
            url = new URL(urlString);
        }
        return url;
    }


    /**
     * Returns a set containing all templates defined in this JAR file.
     */
    private Map<String, String> getJarTemplates(URL url) throws Exception {
        PluginLog.getLog().debug("retrieving all templates from jar file: " + url);
        String lookFor = DIR_TEMPLATES + "/";
        int length = lookFor.length();
        Map<String, String> templates = new HashMap<String, String>();
        BufferedInputStream bis = new BufferedInputStream(url.openStream());
        JarInputStream jis = new JarInputStream(bis);
        JarEntry je;
        Set<String> temp = new HashSet<String>();
        while ((je = jis.getNextJarEntry()) != null) {
            // find the template name
            String jarEntryName = je.getName();
            PluginLog.getLog().debug("Found entry: " + jarEntryName);
            if (jarEntryName.length() <= length || !jarEntryName.startsWith(lookFor)) {
                continue;
            }
            int nextSlashIndex = jarEntryName.indexOf("/", length);
            if (nextSlashIndex == -1) {
                continue;
            }
            String jarTemplate = jarEntryName.substring(length, nextSlashIndex);
            PluginLog.getLog().debug("Found template: " + jarTemplate);
            if (templates.containsKey(jarTemplate)) {
                continue;
            }
            if (jarEntryName.endsWith("readme.txt")) {
                // a description found - add to templates
                String description = getShortDescription(jis);
                templates.put(jarTemplate, description);
                // remove from temp
                temp.remove(jarTemplate);
            } else {
                // add to temp until a description is found
                temp.add(jarTemplate);
            }
        }
        // add all templates that has no description
        for (String aTemp : temp) {
            templates.put(aTemp, "No description found.");
        }
        return templates;
    }


    private String getShortDescription(JarInputStream jis) throws IOException {
        BufferedReader reader = new BufferedReader(new InputStreamReader(jis));
        StringBuilder sb = new StringBuilder();
        boolean first = true;
        while (true) {
            String line = reader.readLine();
            if (line == null || !StringUtils.hasText(line)) {
                break;
            }
            if (!first) {
                sb.append("\n\r");
            } else {
                first = false;
            }
            sb.append(line);
        }
        jis.closeEntry();
        return sb.toString();
    }


    /**
     * Returns an array of available project templates names.
     */
    private Map<String, String> getAvailableTemplates() throws Exception {
        Map<String, String> templates = new HashMap<String, String>();
        Enumeration urls = Thread.currentThread().getContextClassLoader().getResources(DIR_TEMPLATES);
        while (urls.hasMoreElements()) {
            URL url = (URL) urls.nextElement();
            PluginLog.getLog().debug("retrieving all templates from url: " + url);
            Map<String, String> jarTemplates = getJarTemplates(getJarURLFromURL(url, ""));
            templates.putAll(jarTemplates);
        }
        Map<String, String> sortedTemplates = new LinkedHashMap<String, String>();
        String desc = templates.remove("event-processing");
        if (desc != null) {
            sortedTemplates.put("event-processing", desc);
        }
        desc = templates.remove("persistent-event-processing");
        if (desc != null) {
            sortedTemplates.put("persistent-event-processing", desc);
        }
        sortedTemplates.putAll(templates);
        return sortedTemplates;
    }


    /**
     * Creates a String message showing all available templates.
     *
     * @return a String message showing all available templates
     */
    private String createAvailableTemplatesMessage() throws Exception {
        Map<String, String> availableTemplates = getAvailableTemplates();
        StringBuilder sb = new StringBuilder();
        sb.append("\n\n\n\n\n\n");
        if (template == null || template.trim().length() == 0) {
            sb.append("\nPlease use the -Dtemplate=<template> argument to specify a project template.\n");
        } else {
            sb.append("\nThe template '");
            sb.append(template);
            sb.append("' was not found.\n");
        }
        sb.append("Available templates:\n");
        sb.append("--------------------\n");
        Iterator iter = availableTemplates.keySet().iterator();
        int i = 1;
        while (iter.hasNext()) {
            String tmpl = (String) iter.next();
            String desc = availableTemplates.get(tmpl);
            sb.append(i++);
            sb.append(". ");
            sb.append(tmpl);
            sb.append(" - ");
            sb.append(desc);
            sb.append("\n\n");
        }
        return sb.toString();
    }

}